/* $Id: graphics.c,v 1.8 1997/09/15 14:41:33 wtcorrea Exp $ */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "graphics.h"
#include "device.h"
#include "my.h"
#include "macros.h"

int usedev;

void obj_free_data(obj_t *obj)
{
    assert(obj);
    switch (obj->type) {
    case OBJ_LINE:
	line_free(obj->data);
	break;
    case OBJ_POLYLINE:
	polyline_free(obj->data);
	break;
    case OBJ_POLYGON:
	polygon_free(obj->data);
	break;
    case OBJ_TRIANGLE:
	triangle_free(obj->data);
	break;
    case OBJ_RECT:
	rect_free(obj->data);
	break;
    default:
	fprintf(stderr, "obj_free_data: not implemented yet...\n");
	break;
    }
}

void obj_draw(const obj_t *obj)
{
    assert(obj);
    switch (obj->type) {
    case OBJ_LINE:
	line_draw(obj->data, &obj->attr);
	break;
    case OBJ_POLYLINE:
	polyline_draw(obj->data, &obj->attr);
	break;
    case OBJ_POLYGON:
	polygon_draw(obj->data, &obj->attr);
	break;
    case OBJ_TRIANGLE:
	triangle_draw(obj->data, &obj->attr);
	break;
    case OBJ_RECT:
	rect_draw(obj->data, &obj->attr);
	break;
    default:
	fprintf(stderr, "obj_draw: not implemented yet...\n");
	break;
    }
}

void obj_draw_handles(const obj_t *obj)
{
    assert(obj);
    switch (obj->type) {
    case OBJ_LINE:
	line_draw_handles(obj->data);
	break;
    case OBJ_POLYLINE:
	polyline_draw_handles(obj->data);
	break;
    case OBJ_POLYGON:
	polygon_draw_handles(obj->data);
	break;
    case OBJ_TRIANGLE:
	triangle_draw_handles(obj->data);
	break;
    case OBJ_RECT:
	rect_draw_handles(obj->data);
	break;
    default:
	fprintf(stderr, "obj_draw_handles: not implemented yet...\n");
	break;
    }
}

void obj_move(obj_t *obj, int dx, int dy)
{
    assert(obj);
    switch (obj->type) {
    case OBJ_LINE:
	line_move(obj->data, dx, dy);
	break;
    case OBJ_POLYLINE:
	polyline_move(obj->data, dx, dy);
	break;
    case OBJ_POLYGON:
	polygon_move(obj->data, dx, dy);
	break;
    case OBJ_TRIANGLE:
	triangle_move(obj->data, dx, dy);
	break;
    case OBJ_RECT:
	rect_move(obj->data, dx, dy);
	break;
    default:
	fprintf(stderr, "obj_move: not implemented yet...\n");
	break;
    }
}

int obj_pick(const obj_t *obj, int x, int y)
{
    int pick;

    assert(obj);
    switch (obj->type) {
    case OBJ_LINE:
	pick = line_pick(obj->data, x, y);
	break;
    case OBJ_POLYLINE:
	pick = polyline_pick(obj->data, x, y);
	break;
    case OBJ_POLYGON:
	pick = polygon_pick(obj->data, x, y);
	break;
    case OBJ_TRIANGLE:
	pick = triangle_pick(obj->data, x, y);
	break;
    case OBJ_RECT:
	pick = rect_pick(obj->data, x, y);
	break;
    default:
	fprintf(stderr, "obj_pick: not implemented yet...\n");
	break;
    }
    return pick;
}

void gr_disable_device(void)
{
    usedev = 0;
}

void gr_enable_device(void)
{
    usedev = 1;
}

void gr_clear(double r, double g, double b, double a)
{
    if (usedev) {
	dev_clear(r, g, b, a);
    } else {
	my_clear(r, g, b, a);
    }
}

void gr_set_viewport(int x, int y, int w, int h)
{
    if (usedev) {
	dev_set_viewport(x, y, w, h);
    } else {
	my_set_viewport(x, y, w, h);
    }
}

void gr_set_window(double l, double r, double b, double t)
{
    if (usedev) {
	dev_set_window(l, r, b, t);
    } else {
	my_set_window(l, r, b, t);
    }
}

void gr_draw_handle(int x, int y)
{
    attr_t attr = {
	0.0, 0.0, 0.0, 1.0,	/* line color */
	1,			/* line width */
	0, 0, 0x0000,		/* line stipple flag, factor, and pattern */
	1,			/* boundary flag */
	1, 0.0, 0.0, 0.0, 1.0	/* fill flag and color */
    };
    polygon_t *p;

    p = polygon_new(4);
    polygon_addpt(p, x - 2, y - 2);
    polygon_addpt(p, x + 2, y - 2);
    polygon_addpt(p, x + 2, y + 2);
    polygon_addpt(p, x - 2, y + 2);
    polygon_draw(p, &attr);
    polygon_free(p);
}

/* From Graphics Gems p. 49. */
int gr_pt_on_line(int px, int py, int qx, int qy, int tx, int ty)
{
    if (abs((qy - py) * (tx - px) - (ty - py) * (qx - px)) >=
	max(abs(qx - px), abs(qy - py)) + 5) {
	return 0;
    }
    if ((qx < px && px < tx) || (qy < py && py < ty)) {
	return 1;
    }
    if ((tx < px && px < qx) || (ty < py && py < qy)) {
	return 1;
    }
    if ((px < qx && qx < tx) || (py < qy && qy < ty)) {
	return 3;
    }
    if ((tx < qx && qx < px) || (ty < qy && qy < py)) {
	return 3;
    }
    return 2;
}

int gr_pt_in_triang(vertex_t v0, vertex_t v1, vertex_t v2, int x, int y)
{
    int v0p[2], v1p[2], v2p[2], v0v1[2], v1v2[2], v2v0[2];
    int a, b, c;

    v0p[0] = x - v0.x, v0p[1] = y - v0.y;
    v1p[0] = x - v1.x, v1p[1] = y - v1.y;
    v2p[0] = x - v2.x, v2p[1] = y - v2.y;

    v0v1[0] = v1.x - v0.x, v0v1[1] = v1.y - v0.y;
    v1v2[0] = v2.x - v1.x, v1v2[1] = v2.y - v1.y;
    v2v0[0] = v0.x - v2.x, v2v0[1] = v0.y - v2.y;

    a = sign(v0p[0] * v0v1[1] - v0p[1] * v0v1[0]);
    b = sign(v1p[0] * v1v2[1] - v1p[1] * v1v2[0]);
    c = sign(v2p[0] * v2v0[1] - v2p[1] * v2v0[0]);
    return (a == b) && (b == c);
}
